#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include "utils.h"
#include "eigen.h"

#define SQR(a) (((a) == 0.0) ? 0.0 : (a) * (a))
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

#define toler 1e-9

#define _DEBUG 1

/* reduce n x n symmetric matrix to tridiagonal using Givens method */
void tridiag(double *m, int n, double *d, double *e);
inline void givens(double* g, double a, double b);

/* use QL algorithm with implicit shifts to compute eigenvecors and eigenvalues */
void tqli(double *d, double *e, int n, double *z);

void resort(double *d, double *a, int n);


/* compute eigen vectors and eigen values of a symmetric matrix 
   return orthonormal eigenvectors of the matrix in Matrix and
   eigen values in eigenvalues */

void eigen(double *Matrix, int dim, double *eigenvalues)
{
    double *e;
    
#ifdef _DEBUG
    struct timeval t0, t1;
#endif
    
    if ((e = (double *)malloc(dim * sizeof(double))) == NULL)
    {
	printf("ERROR: malloc failed in eigen for vector size %i\n", dim);
	exit(1);
    }

#ifdef _DEBUG
{
    printf("Running tridiagonalization...\n");
    gettimeofday(&t0, NULL);
}
#endif 
	
    /* reduce n x n symmetric matrix to tridiagonal using Givens method */
    tridiag(Matrix, dim, eigenvalues, e); 
    
#ifdef _DEBUG
{
    int i;
    
    gettimeofday(&t1, NULL);
    printf("Done tridiagonalization in %.2f sec\n", TDIFF(t1, t0));
    printf("Running tqli...\n");
    gettimeofday(&t0, NULL);
}
#endif

    /* use QL algorithm with implicit shifts to compute eigenvecors and eigenvalues */
    tqli(eigenvalues, e, dim, Matrix);

#ifdef _DEBUG    
{
    gettimeofday(&t1, NULL);
    printf("Done tqli in %.2f sec\n", TDIFF(t1, t0));
}
#endif

    /* rearrange the values in ascending order */
    resort(eigenvalues, Matrix, dim);

    free(e);
}


inline void apply_givens_2_rows(double *m, int n, double *g, int r1, int r2, int j)
{
    int ii;
    
    for (ii = j; ii < n; ii++) 
    {
        int i_ii_addr = r2 * n + ii;
        int j1_ii_addr = r1 * n + ii;
		
        double a =  g[0] * m[j1_ii_addr] + g[1] * m[i_ii_addr];
        double b =  g[2] * m[j1_ii_addr] + g[3] * m[i_ii_addr];
	
        m[j1_ii_addr] = a;
        m[i_ii_addr]  = b;
    }
}

inline void apply_givens_2_cols(double *m, int n, double *g, int c1, int c2, int j)
{
    int jj;
    
    for (jj = j; jj < n; jj++)
    {
        int jj_j1_addr = jj * n + c1;
        int jj_i_addr = jj * n + c2;
	    
        double a = m[jj_j1_addr] * g[0] + m[jj_i_addr] * g[1];
        double b = m[jj_j1_addr] * g[2] + m[jj_i_addr] * g[3];
	
        m[jj_j1_addr] = a;
        m[jj_i_addr] = b;
    }
}


void tridiag(double *m, int n, double *d, double *e)
{
    int i, j, ii, jj, k=0, p;
    double a, b, c;
    int i_ii_addr, j1_ii_addr, jj_j1_addr, jj_i_addr;
    double g[4];
    double g_stored[n][4];

    double *x = (double *)malloc(n* n * sizeof(double));
    if (x == NULL)
    	exit (-1);

    bzero(x, n * n * sizeof(double));
    for (i = 0; i < n; i++)
        x[i * n + i] = 1;

    int step;

    for (j = 0; j < n - 2; j++) 
    {
        for (i = j + 2; i < n; i++)
	{
            givens(g_stored[i], m[(j+1)*n + j], m[i*n + j]);
            apply_givens_2_rows(m, n, g_stored[i], j+1, i, j);
	}

        for (i = j + 2; i < n; i++)
            apply_givens_2_cols(m, n, g_stored[i], j+1, i, j);

        for (i = j + 2; i < n; i++)
            apply_givens_2_cols(x, n, g_stored[i], j+1, i, 0);
    }

    /* copy diagonal and subdiagonal values */
    d[0] = m[0];
    for (i = 1; i < n; i++)
    {
        d[i] = m[i*n+i];
        e[i] = m[i*n+i-1];
    }
	
    /* copy x into m */
    memcpy(m, x, n * n * sizeof(double));
    
    free(x);
}


inline void givens(double* g, double a, double b)
{
    double r = sqrt(a*a + b*b);
    double rr = 1.0 / r;
    double c = a * rr;
    double s = b * rr;

    g[0] =  c;
    g[1] =  s;
    g[2] = -s;
    g[3] =  c;
}

/* use QL algorithm with implicit shifts to compute eigenvecors and eigenvalues */
void tqli(double *d, double *e, int n, double *z)
{
    int m, l, iter, i, k;
    double s, r, p, g, f, dd, c, b;
    int k_addr;

    for (i = 1; i < n; i++)
        e[i-1] = e[i];
    
    e[n-1] = 0.0; 
	
    for (l = 0; l < n; l++) 
    {
        iter = 0;
        do 
        {
            for (m = l; m < n-1; m++) 
            {
                dd = fabs(d[m]) + fabs(d[m+1]);
                //if ((fabs(e[m]) + dd) == dd) break;
		if (fabs(e[m]) < toler) break;
            }

            if (m != l)
            {
                if (iter++ == 30)
                    printf("Too many iterations in tqli\n");
		    
                g = (d[l+1] - d[l]) / (2.0 * e[l]);
                //r = pythag(g, 1.0);
		r = sqrt(g * g + 1.0);
                g = d[m] - d[l] + e[l] / (g + SIGN(r,g));
                s = c = 1.0;
                p = 0.0;

                for (i = m-1; i >= l; i--) 
                {
                    f = s * e[i];
                    b = c * e[i];
		    
                    if (fabs(f) >= fabs(g)) 
                    {
                        c = g / f;
                        r = sqrt(c * c + 1.0);
                        e[i+1] = f * r;
                        s = 1.0 / r;
                        c *= s;
                    }
                    else 
                    {
                        s = f / g;
                        r = sqrt(s * s + 1.0);
                        e[i+1] = g * r;
                        c = 1.0 / r;
                        s *= c;
                    }

                    g = d[i+1] - p;
                    r = (d[i] - g) * s + 2.0 * c * b;
		    p = s * r;
                    d[i+1] = g + p;
                    g = c * r - b;

                    for (k = 0; k < n; k++) 
                    {
                        k_addr = k * n;
                        f = z[k_addr+i+1]; 
                        z[k_addr+i+1] = s * z[k_addr+i] + c * f; 
                        z[k_addr+i] = c * z[k_addr+i] - s * f; 
                    }
                }

                if (r == 0.0 && i >= l) continue;

                d[l] -= p;
                e[l] = g;
                e[m] = 0.0;
            }
        } while (m != l);
    }
}


void resort(double *d, double *a, int n)
{
    int i, j, k, min_indx;
    double tmp;
    
    /* first, transpose the eigenvector matrix */
    for (i = 0; i < n-1; i++)
        for (j = i+1; j < n; j++)
        {
            tmp = a[i*n+j];
            a[i*n+j] = a[j*n+i];
            a[j*n+i] = tmp;
        }
    
    for (i = 0; i < n-1; i++)
    {
        /* find the smallest value in the list of unsorted values */
        min_indx = i;
        for (j = i+1; j < n; j++)
	{
            if (d[min_indx] > d[j])
                min_indx = j;
	}

        if (min_indx != i)
        {
            /* swap it with first unsorted value */
            tmp = d[i];
            d[i] = d[min_indx];
            d[min_indx] = tmp;
	
            /* swap rows in the matrix */
            for (k = 0; k < n; k++)
            {
                tmp = a[i*n+k];
                a[i*n+k] = a[min_indx*n+k];
                a[min_indx*n+k] = tmp;
            }
        }
    }
}

